#lang scheme
(require "test.ss")

;; Combinators

;; ----------------------------------------

;; no-more : next
(define (no-more success-k fail-k)
  (fail-k))

;; or-k : next next -> (success fail -> alpha)
(define ((or-k first second) success-k fail-k)
  (local [(define (make-fst-success-k success-k fail-k)
            (lambda (val stream new-next)
              (success-k val stream
                         (lambda (success-k fail-k)
                           (new-next (make-fst-success-k success-k fail-k)
                                     (make-fst-fail-k success-k fail-k))))))
          (define (make-fst-fail-k success-k fail-k)
            (lambda ()
              (second success-k fail-k)))]
    (first (make-fst-success-k success-k fail-k)
           (make-fst-fail-k success-k fail-k))))

;; and-k : next (stream -> next) -> (success fail -> alpha)
(define ((and-k first second combine) success-k fail-k)
  (local [(define (make-fst-success-k success-k fail-k)
            (lambda (fst-vals stream fst-next)
              ((second stream)
               (make-snd-success-k fst-vals fst-next success-k fail-k)
               (lambda ()
                 (fst-next (make-fst-success-k success-k fail-k)
                           fail-k)))))
          (define (make-snd-success-k fst-vals fst-next success-k fail-k)
            (lambda (snd-vals stream snd-next)
              (success-k 
               (combine fst-vals snd-vals)
               stream
               (lambda (success-k fail-k)
                 (snd-next 
                  (make-snd-success-k fst-vals fst-next success-k fail-k)
                  (lambda ()
                    (fst-next (make-fst-success-k success-k fail-k)
                              fail-k)))))))]
    (first (make-fst-success-k success-k fail-k)
           fail-k)))

;; wrap-success-k : next (any -> any) -> success fail -> allpha
(define (wrap-success-k try val->val)
  (local [(define (make-my-success-k success-k fail-k)
            (lambda (val stream next)
              (success-k (val->val val) stream
                         (lambda (success-k fail-k)
                           (next (make-my-success-k success-k fail-k)
                                 fail-k)))))]
    (lambda (success-k fail-k)
      (try (make-my-success-k success-k fail-k) fail-k))))

;; ----------------------------------------

(define table
  `([many [(".") ,(lambda () null)]
          [(main many) ,cons]]
    [main [("apple") ,(lambda () 'apple)]
          [("banana") ,(lambda () 'banana)]
          [(fruit-salad "!") ,(lambda (x)
                                        `(,x !))]]
    [fruit-salad [(fruit fruit) ,(lambda (x y)
                                   (list x y))]
                 [(fruit fruit fruit) ,(lambda (x y z)
                                         (list x y z))]]
    [fruit [("orange") ,(lambda () 'orange)]
           [("grape") ,(lambda () 'grape)]]))

;; A ptable is list-of-(cons symbol box-of-(list-of-string -> next))

;; make-ptable : table -> ptable
(define (make-ptable table)
  (let ([ptable (map (lambda (defn) 
                       (cons (car defn) (box #f)))
                     table)])
    (for-each (lambda (defn pdefn)
                (set-box! (cdr pdefn)
                          (parse-alts (cdr defn) ptable)))
              table ptable)
    ptable))

;; parse : sym ptable -> list-of-string -> next
(define (parse what ptable)
  (local [(define pdefn (assq what ptable))]
    (if pdefn
        (let ([b (cdr pdefn)])
          (lambda (stream) ((unbox b) stream)))
        (error 'parse "no definition for ~a" what))))

(define (ok v stream next) (cons v stream))
(define (nope) #f)

(test (((parse 'main ptable) '("apple")) ok nope) (cons 'apple '()))
(test (((parse 'fruit-salad ptable) '("orange" "orange")) ok nope) 
      (cons '(orange orange) '()))
(test (((parse 'main ptable) '("orange" "orange" "!")) ok nope)
      (cons '((orange orange) !) '()))
(test (((parse 'main ptable) '("orange" "orange" "orange" "!")) ok nope)
      (cons '((orange orange orange) !) '()))
(test (((parse 'main ptable) '("radish")) ok nope) #f)


;; parse-alts : list-of-case ptable -> list-of-string -> next
(define (parse-alts alts ptable)
  (if (null? alts)
      (lambda (stream) no-more)
      (let ([f (parse-case (first alts) ptable)]
            [r (parse-alts (rest alts) ptable)])
        (lambda (stream)
          (or-k (f stream) (r stream))))))

(let ([alts (list
             `[("apple") ,(lambda () 'apple)]
             `[("grape") ,(lambda () 'grape)])])
  (test (((parse-alts alts ptable) '("apple")) ok nope) (cons 'apple '()))
  (test (((parse-alts alts ptable) '("grape")) ok nope) (cons 'grape '()))
  (test (((parse-alts alts ptable) '("grape" "fish")) ok nope)
        (cons 'grape '("fish")))
  (test (((parse-alts alts ptable) '("radish")) ok nope) #f))

;; parse-case : case ptable -> list-of-string -> next
(define (parse-case case ptable)
  (local [(define seq (first case))
          (define handler (second case))]
    (let ([s (parse-sequence seq ptable)])
      (lambda (stream)
        (wrap-success-k
         (s stream)
         (lambda (vals) (apply handler vals)))))))

(let ([case `[("apple") ,(lambda () 'apple)]])
  (test (((parse-case case ptable) '("apple")) ok nope) (cons 'apple '()))
  (test (((parse-case case ptable) '("grape")) ok nope) #f))


;; parse-sequence : sequence ptable -> list-of-string -> next
(define (parse-sequence seq ptable)
  (cond
    [(empty? seq) 
     (lambda (stream)
       (lambda (success-k fail-k)
         (success-k empty stream no-more)))]
    [else 
     (let ([f (parse-item (first seq) ptable)]
           [r (parse-sequence (rest seq) ptable)])
       (lambda (stream)
         (and-k (f stream)
                r
                append)))]))

(let ([seq `("apple" "grape")])
  (test (((parse-sequence '("apple") ptable) '("apple")) ok nope) 
        (cons empty '()))
  (test (((parse-sequence seq ptable) '("apple" "grape")) ok nope)
        (cons empty '()))
  (test (((parse-sequence seq ptable) '("apple" "radish")) ok nope) #f))
(let ([seq '(fruit fruit)])
  (test (((parse-sequence seq ptable) '("orange" "grape")) ok nope)
        (cons '(orange grape) '()))
  (test (((parse-sequence seq ptable) '("grape" "orange")) ok nope)
        (cons '(grape orange) '())))
(let ([seq '(fruit-salad "!")])
  (test (((parse-sequence seq ptable) '("orange" "orange" "!")) ok nope)
        (cons '((orange orange)) '()))
  (test (((parse-sequence seq ptable) '("orange" "orange" "orange" "!")) ok nope)
        (cons '((orange orange orange)) '())))


;; parse-item : sequence ptable -> list-of-string -> next
(define (parse-item item ptable)
  (cond
    [(string? item)
     (lambda (stream)
       (if (and (cons? stream)
                (equal? item (first stream)))
           (lambda (success-k fail-k)
             (success-k empty (rest stream) no-more))
           no-more))]
    [else
     (let ([nt (parse item ptable)])
       (lambda (stream)
         (wrap-success-k (nt stream) list)))]))

(test (((parse-item '"grape" ptable) '("grape")) ok nope)
      (cons empty '()))
(test (((parse-item 'fruit ptable) '("grape")) ok nope)
      (cons '(grape) '()))
(test (((parse-item 'fruit ptable) '("radish")) ok nope) #f)

(define ptable (make-ptable table))

(run-tests)

(time
 (for ([i (in-range 10000)])
   (let ([p (parse 'many ptable)])
     ((p '("orange" "orange" "orange" "!" 
                    "grape" "orange" "!"
                    "orange" "orange" "grape" "!"
                    ".")) 
      ok nope))))
